package Net::Hotline;

# Provides the version number for the whole set of Net::Hotline:: libs.
# Currently, this includes only the Net::Hotline::Client
$VERSION = '0.81';

# The client module is loaded when you "use Net::Hotline"
use Net::Hotline::Client;

1;
